# Generated by Django 5.2.9 on 2026-02-19 10:23
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0028_fix_device_virtualchassis_removal"),
        ("users", "0015_owner"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricsource",
            name="owner",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="users.owner",
            ),
        ),
    ]
