{% load i18n %}
<div class="row mb-3">
    <div class="col-12">
        <div class="card">
            <h5 class="card-header">
                {% if object %}
                    {% if context_type == "endpoint" or context_type == "endpoint_all" %}
                        {% trans "Combined Filters" %}: {{ object.endpoint }}
                        {% if sync %}
                            <small class="text-muted">&nbsp;- {% trans "Sync:" %} {{ sync.name }} (ID: {{ sync.pk }})</small>
                        {% else %}
                            <small class="text-muted">&nbsp;- {% trans "All Syncs" %}</small>
                        {% endif %}
                    {% else %}
                        {% trans "Merged Filter Expressions for filter type" %}: {{ object.filter_type|upper }}
                    {% endif %}
                {% else %}
                    {% trans "Combined Filters" %}
                {% endif %}
            </h5>
            <div class="card-body">
                {% if is_empty %}
                    <div class="alert alert-info" role="alert">
                        <i class="mdi mdi-information"></i>
                        {% if context_type == "endpoint" %}
                            {% trans "No filters are applied to this endpoint in this sync." %}
                        {% elif context_type == "endpoint_all" %}
                            {% trans "No filters are using this endpoint." %}
                        {% else %}
                            {% trans "No expressions defined for this filter." %}
                        {% endif %}
                    </div>
                {% else %}
                    <pre>{{ combined_expressions | json }}</pre>
                {% endif %}
            </div>
        </div>
    </div>
</div>
