# Generated by Django 5.2.5 on 2025-12-02 12:06
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("extras", "0132_configcontextprofile"),
        ("ipfabric_netbox", "0023_populate_filters_data"),
    ]

    operations = [
        # The new endpoint is populated, so we can make the field non-nullable
        migrations.AlterField(
            model_name="ipfabrictransformmap",
            name="source_endpoint",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="transform_maps",
                to="ipfabric_netbox.ipfabricendpoint",
            ),
        ),
        # Remove the old source_model field as it's been replaced by source_endpoint
        migrations.RemoveField(
            model_name="ipfabrictransformmap",
            name="source_model",
        ),
    ]
